
from . import utilities
from Bio.PDB import PDBParser
from Bio.PDB.ResidueDepth import _get_atom_radius
import glob
import itertools
import numpy as np
import os
import pandas as pd
import scipy
import shutil
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import normalize
import sys
import tempfile
import trimesh


class Spheres(object):

    def __init__(self,
                 xyz=None,
                 r=None,
                 xyzr=None,
                 xyzrg=None,
                 g=None,
                 pdb=None,
                 bv=None,
                 mesh=None,
                 name=None):
        """
        A Spheres object contains a list of xyz centers with r radii and g groups. It can be defined using xyzrg, xyzr (and optionally g), xyz (and optionally r or g), a pdb file (and optionally r or g), or a list of vertices with normals bounded by the spheres (requires r and optionally includes g)

        Parameters
        ----------
        xyz : (n, 3) float
            Array containing centers
        r : (n, 1) float
            Array containing radii
        xyzr : (n, 4) float
            Array containing centers and radii
        xyzrg : (n, 5) float
            Array containing centers, radii, and groups
        g : (n, 1) float
            Array containing groups
        pdb : string
            filename of a pdb to be processed into spheres
        bv : (n, 6) float
            Array containing vertices and normals
        """

        if xyzrg is not None:
            self.xyzrg = xyzrg
        elif xyzr is not None:
            self.xyzr = xyzr

            if g is not None:
                self.g = g
        elif xyz is not None:
            self.xyz = xyz

            if r is not None:
                self.r = r
            if g is not None:
                self.g = g
        elif pdb is not None:
            if not sys.warnoptions:
                import warnings
                warnings.simplefilter("ignore")

            p = PDBParser(PERMISSIVE=1, QUIET=True)
            structure = p.get_structure("prot", pdb)

            self.xyz = np.array([atom.get_coord() for atom in structure[0].get_atoms()])

            if r is not None:
                self.r = r
            else:
                self.r = [_get_atom_radius(atom, rtype='united') for atom in structure[0].get_atoms()]

            if g is not None:
                self.g = g
        elif bv is not None and r is not None:
            self.xyz = bv[:, 0:3] + r * bv[:, 3:6]
            self.r = r
            self.remove_duplicates()

            if g is not None:
                self.g = g
        if mesh is not None:
            self.mesh = mesh
        else:
            self.mesh = None

        if name is not None:
            self.name = name
        else:
            self.name = None


    def __add__(self, other):
        """
        Create a new Spheres object by overloading addition to concatenate xyzr contents
        Does not add meshes (just spheres)
        """
        
        if other is not None:
            return Spheres(xyzrg=np.concatenate([self.xyzrg, other.xyzrg], axis=0))
        else:
            return Spheres(xyzrg=np.copy(self.xyzrg))


    def copy(self):
        return Spheres(xyzrg=np.copy(self.xyzrg))


    def calculate_surface(self, probe_radius=1.4, cavity_atom=None, coordinate=None, all_components=False, exclusionary_radius=2.5, largest_only=False, noh=True, minimum_volume=200):
        """
        Calculate the SAS for a given probe radius 

        Parameters
        ----------
        probe_radius : float
            probe radius for msms calculation
        cavity_atom : int
            index of an atom used to define the binding pocket surface; only used for interior cavity identification; exclusive of all_components
        all_components : boolean
            flag to calculate all components; output mesh is None
        exclusionary_radius : float
            maximum radius of atoms considered by the nearest function (excludes pseudoatoms from identification)
        """

        tmpdir = tempfile.mkdtemp()
        xyzr_file = os.path.join(tmpdir, "pyvol.xyzr")
        msms_template = os.path.join(tmpdir, "pyvol_msms")
        
        np.savetxt(xyzr_file, self.xyzr, delimiter=' ')
        if (cavity_atom is None) and (coordinate is not None):
            cavity_atom = self.nearest(coordinate, max_radius=exclusionary_radius)
        
        msms_cmd = ["msms", "-if", xyzr_file, "-of", msms_template, "-probe_radius", "{0}".format(probe_radius), "-no_area"]
        if noh:
            msms_cmd.append("-noh")
        if cavity_atom is not None:
            msms_cmd.extend(["-one_cavity", 1, cavity_atom])
        elif all_components:
            msms_cmd.append("-all_components")

        utilities.run_cmd(msms_cmd)

        sphere_list = []

        def read_msms_output(msms_template):
            try:
                verts_raw = pd.read_csv("{0}.vert".format(msms_template), sep='\s+', skiprows=3, dtype=np.float_, header=None, encoding='latin1').values
                faces = pd.read_csv("{0}.face".format(msms_template), sep='\s+', skiprows=3, usecols=[0, 1, 2], dtype=np.int_, header=None, encoding='latin1').values
            except IOError:
                print("MSMS failed to run correctly", msms_template)
                raise IOError
            else:
                vertices = np.zeros((verts_raw.shape[0] + 1, 3))
                vertices[1:, :] = verts_raw[:, 0:3]
                return verts_raw, vertices, faces

        if not all_components:
            verts_raw, vertices, faces = read_msms_output(msms_template)

            mesh = trimesh.base.Trimesh(vertices=vertices, faces=faces)
            if mesh.volume < 0:
                faces = np.flip(faces, axis=1)
                mesh = trimesh.base.Trimesh(vertices=vertices, faces=faces)
            bspheres = Spheres(bv=verts_raw, r=probe_radius, mesh=mesh)
            shutil.rmtree(tmpdir)
            return [bspheres]
            
        else:
            spheres_list = []
            ac_template_list = [os.path.splitext(x)[0] for x in glob.glob("{0}_*.face".format(msms_template))]

            largest_mesh = None
            for ac_template in ac_template_list:
                verts_raw, vertices, faces = read_msms_output(ac_template)
                
                tm = trimesh.base.Trimesh(vertices=vertices, faces=faces)
                if tm.volume < 0:
                    tm = trimesh.base.Trimesh(vertices=vertices, faces=np.flip(faces, axis=1))
                
                if largest_only:
                    if largest_mesh is None:
                        largest_mesh = tm
                        bspheres = Spheres(bv=verts_raw, r=probe_radius, mesh=tm)
                    elif tm.volume > largest_mesh.volume:
                        largest_mesh = tm
                        bspheres = Spheres(bv=verts_raw, r=probe_radius, mesh=tm)
                else:
                    if tm.volume >= minimum_volume:
                        bspheres = Spheres(bv=verts_raw, r=probe_radius, mesh=tm)
                        spheres_list.append(bspheres)

            if largest_only:
                return [bspheres]
            else:
                return spheres_list
            
    
    def identify_nonextraneous(self, ref_spheres, radius):
        """
        Returns all spheres less than radius away from any center in ref_spheres using cKDTree search built on the non-reference set

        Parameters
        ----------
        ref_spheres : Spheres
            A Spheres object containing the reference set
        radius : float
            Maximum distance between sets for inclusion
        """

        kdtree = scipy.spatial.cKDTree(self.xyz)
        groups = kdtree.query_ball_point(ref_spheres.xyz, radius, n_jobs=-1)
        indices = np.unique(list(itertools.chain.from_iterable(groups)))
        
        return Spheres(xyzrg=np.copy(self.xyzrg[indices, :]))


    def nearest(self, coordinate, max_radius=None):
        """
        returns the index of the sphere closest to a coordinate; if max_radius is specified, the sphere returned must have a radius <= max_radius

        Parameters
        ----------
        coordinate : (1, 3) float
            3D input coordinate
        max_radius : float
            maximum radius that the selected nearest sphere can have; used to exclude large radius exterior spheres from being selected 
        """

        if max_radius is None:
            sphere_list = self.xyz
        else:
            sphere_list = self.xyz[self.r <= max_radius]

        return np.argmin(scipy.spatial.distance.cdist(sphere_list, coordinate))


    def nearest_coord_to_external(self, coordinates):
        """
        returns the coordinate of the sphere closest to the supplied coordinates

        Parameters
        ----------
        coordinates : (n, 3) float
            3D input coordinates
        """
 
        kdtree = scipy.spatial.cKDTree(self.xyz)
        dist, indices = kdtree.query(coordinates, n_jobs=-1)
        
        return self.xyz[indices[np.argmin(dist)], :]
    
                
    def remove_duplicates(self, eps=0.01):
        """
        Remove duplicate spheres by identifying centers closer together than eps using DBSCAN

        Parameters
        ----------
        eps : float
            DBSCAN cutoff for identifying nearest neighbor distances between duplicate spheres
        """
        
        db = DBSCAN(eps=eps, min_samples=1).fit(self.xyz)
        values, indices = np.unique(db.labels_, return_index=True)
        self.xyzrg = self.xyzrg[indices, :]


    def remove_ungrouped(self):
        ungrouped_indices = np.where(self.g < 1)
        self.xyzrg = np.delete(self.xyzrg, ungrouped_indices, axis=0)
        self.mesh = None
        
        
    def write(self, filename, contents="xyzr", output_mesh=None):
        """
        Writes the contents of _xyzrg to a space delimited file 

        Parameters
        ----------
        filename : string
            filename for the output
        """

        if contents == "xyzrg":
            np.savetxt(filename, self.xyzrg, delimiter=' ')
        elif contents == "xyzr":
            np.savetxt(filename, self.xyzr, delimiter=' ')
        elif contents == "xyz":
            np.savetxt(filename, self.xyz, delimiter=' ')

        if output_mesh is not None:
            if self.mesh is None:
                print("Cannot write out an uninitialized mesh")
            else:
                if output_mesh == " ":
                    output_mesh = "{0}.obj".format(os.path.splitext(filename)[0])
                mesh.export(output_mesh)


    @property
    def xyzrg(self):
        return self._xyzrg


    @xyzrg.setter
    def xyzrg(self, value):
        if value.shape[1] != 5:
            raise ValueError("number of xyzrg array columns must equal 5")
        self._xyzrg = np.copy(value).astype(float)

        
    @property
    def xyzr(self):
        return self._xyzrg[:, 0:4]

    
    @xyzr.setter
    def xyzr(self, value):
        # resets all radii, groups, and positions
        if value.shape[1] != 4:
            raise ValueError("number of xyzr array columns must equal 4")
        xyzrg = np.zeros((value.shape[0], 5))
        xyzrg[:, 0:4] = value
        self._xyzrg = np.copy(xyzrg).astype(float)

        
    @property
    def xyz(self):
        return self._xyzrg[:, 0:3]

    
    @xyz.setter
    def xyz(self, value):
        # resets all radii, groups, and positions
        if value.shape[1] != 3:
            raise ValueError("number of xyz array columns must equal 3")
        xyzrg = np.zeros((value.shape[0], 5))
        xyzrg[:, 0:3] = value
        self._xyzrg = np.copy(xyzrg).astype(float)

        
    @property
    def r(self):
        return self._xyzrg[:, 3]

    
    @r.setter
    def r(self, value):
        if value is np.ndarray:
            if self._xyzrg.shape[0] == value.shape[0]:
                self._xyzrg[:, 3] = np.copy(value).astype(float)
            else:
                raise ValueError("Number of radii values must match the number of rows in the internal xyz array")
        else:
            self._xyzrg[:, 3] = value


    @property
    def g(self):
        return self._xyzrg[:, 4]

    
    @g.setter
    def g(self, value):
        if value is np.ndarray:
            if self._xyzrg.shape[0] == value.shape[0]:
                self._xyzrg[:, 4] = np.copy(value).astype(float)
            else:
                raise ValueError("Number of group values must match the number of rows in the internal xyzr array")
        else:
            self._xyzrg[:, 4] = value.astype(float)




from .spheres import Spheres
from . import cluster, utilities
import itertools
import numpy as np


def pocket(prot_file, mode="largest", lig_file=None, coordinate=None, residue=None, min_rad=1.4, max_rad=3.4, lig_excl_rad=None, lig_incl_rad=None, subdivide=False, minimum_volume=200, min_subpocket_rad=1.7, max_clusters=None, prefix="pocket", output_dir=None):
    """
    Calculates the SES for a binding pocket

    Parameters
    ----------
    prot_file : string
        PDB filename for the structure containing all atoms which should be considered boundaries of the binding pocket
    mode : string
        Specifies whether to calculate 'all' pockets, the 'largest' pocket, or a 'specific' pocket
    lig_file : string
        PDB filename for the structure containing all atoms of the ligand
    cavity_coordinate : (n, 3) float
        The 3D coordinate used to identify the binding pocket; the nearest atom in the protein structure is assumed to be on the surface of the binding pocket
    min_rad : float
        The minimum radius used to calculate the boundaries of the binding pocket; defaults to 1.7 (the radius of methyl); smaller radii can identify connected subpockets, but these are necessarily inaccesable to small molecules in the absence of protein movement
    max_rad : float
        The maximum radius used to identify the solvent exposed exterior surface of the protein; defaults to 3.4 (approximately the radius of a free alanine)
    lig_excl_rad : float
        Sets an outer bounds to identified pockets at a fixed radius from a provided ligand; provides a convenient method to select a subset of a binding pocket that is not geometrically obvious or unique
    lig_incl_rad : float
        Sets an inner bounds to the solvent exposed exterior surface of the protein at a fixed radius from a provided ligand; allows the user to include a volume of solvent exposed surface in volume calculations when a ligand is known to extend into solvent; this can maintain comparability between ligand and binding pocket volume calculations
    """

    min_rad = float(min_rad)
    max_rad = float(max_rad)

    if min_rad < 1.2:
        print("Warning: minimum radii under 1.2 and not supported and can lead to bizarre results or crashes; setting the minimum radius to 1.2")
        min_rad = 1.2
    
    p_s = Spheres(pdb=prot_file)

    if lig_file is not None:
        if lig_incl_rad is not None:
            lig_incl_rad = float(lig_incl_rad)
        l_s = Spheres(pdb=lig_file, r=lig_incl_rad)
    else:
        l_s = None

    pl_s = p_s + l_s
        
    pl_bs = pl_s.calculate_surface(probe_radius=max_rad)[0]

    pa_s = p_s + pl_bs
    if (l_s is not None) and (lig_excl_rad is not None):
        le_s = Spheres(xyz=l_s.xyzr, r=float(lig_excl_rad))
        le_bs = le_s.calculate_surface(probe_radius=max_rad)[0]
        pa_s = pa_s + le_bs

    if mode == "all":
        all_pockets = pa_s.calculate_surface(probe_radius=min_rad, all_components=True, minimum_volume=minimum_volume)
        for index, pocket in enumerate(all_pockets):
            pocket.name = "{0}_p{1}".format(prefix, index)
    else:
        if mode == "largest":
            bp_bs = pa_s.calculate_surface(probe_radius=min_rad, all_components=True, largest_only=True)[0]
        elif mode == "specific":
            if coordinate is not None:
                coordinate = coordinate.reshape(1, -1)                
            elif residue is not None:
                residue = str(residue)
                chain = None
                if not residue[0].isdigit():
                    chain = residue[0]
                    residue = int(residue[1:])
                else:
                    residue = int(residue)
                res_coords = utilities.coordinates_for_residue(prot_file, residue=residue, chain=chain)
                p_bs = p_s.calculate_surface(probe_radius=min_rad)[0]
                coordinate = p_bs.nearest_coord_to_external(res_coords).reshape(1, -1)
            elif l_s is not None:
                lig_coords = l_s.xyz
                coordinate = np.mean(l_s.xyz, axis=0).reshape(1, -1)
            else:
                print("A coordinate, ligand, or residue must be supplied to run in specific mode")
                return None
            bp_bs = pa_s.calculate_surface(probe_radius=min_rad, coordinate=coordinate)[0]
        else:
            print("Error: unrecognized mode <{0}>--should be 'all', 'largest', or 'specific'".format(mode))

        bp_bs.name = "{0}_p0".format(prefix)

        if bp_bs.mesh.volume > pl_bs.mesh.volume:
            print("Error: binding pocket not correctly identified--try an alternative method to specify the binding pocket")
            return None
        else:
            all_pockets = [bp_bs]
            
        if subdivide:
            all_pockets.extend(subpockets(bounding_spheres = pa_s, ref_spheres = bp_bs, min_rad=min_rad, max_rad=max_rad, min_subpocket_rad=min_subpocket_rad, max_clusters=max_clusters, prefix=prefix))

    if output_dir is not None:
        write_report(all_pockets, output_dir, prefix)
        
    return all_pockets


def subpockets(bounding_spheres, ref_spheres, min_rad, max_rad, min_subpocket_rad=1.7, max_subpocket_rad=None, sampling=0.1, inclusion_radius_buffer=1.0, min_cluster_size=10, max_clusters=None, prefix=None):
    
    if max_subpocket_rad is None:
        max_subpocket_rad = max_rad

    nonextraneous_rad = min_rad + max_rad + inclusion_radius_buffer
    nonextraneous_spheres = bounding_spheres.identify_nonextraneous(ref_spheres=ref_spheres, radius=nonextraneous_rad)

    sampling_radii = np.flip(np.arange(min_rad, max_subpocket_rad, sampling), axis=0)
    unmerged_sphere_lists = utilities.sphere_multiprocessing(nonextraneous_spheres, sampling_radii, all_components=True)
    spheres = cluster.merge_sphere_list(itertools.chain(*unmerged_sphere_lists))

    cluster.hierarchically_cluster_spheres(spheres, ordered_radii=sampling_radii, min_new_radius=min_subpocket_rad, min_cluster_size=min_cluster_size, max_clusters=max_clusters)

    cluster.remove_overlap(spheres, radii=sampling_radii, spacing=sampling)
    cluster.remove_overlap(spheres)
    cluster.remove_interior(spheres)
    grouped_list = cluster.extract_groups(spheres, surf_radius=min_rad, prefix=prefix)
    return grouped_list


def write_report(all_pockets, output_dir, prefix):
    import os
    import pandas as pd

    utilities.check_dir(output_dir)
    
    rept_list = []
    
    for pocket in all_pockets:
        mesh_name = os.path.join(output_dir, "{0}.obj".format(pocket.name))
        pocket.mesh.export(file_obj=mesh_name)        
        rept_list.append({"name": pocket.name,
                          "volume": pocket.mesh.volume
                          })
    rept_df = pd.DataFrame(rept_list)
    rept_name = os.path.join(output_dir, "{0}_rept.csv".format(prefix))
    rept_df.to_csv(rept_name, index=False)




from . import identify
from . import pymol_utilities
import os
from pymol import cgo, cmd, CmdException
import shutil
import tempfile


def pocket(protein, mode=None, ligand=None, pocket_coordinate=None, residue=None, prefix="bp", min_rad=1.4, max_rad=3.4, lig_excl_rad=None, lig_incl_rad=None, display_mode="solid", color='marine', alpha=0.85, output_dir=None, subdivide=None, minimum_volume=200):
    """
    Calculates the SES for a binding pocket and displays it

    Parameters
    ----------
    """

    if output_dir is None:
        out_dir = tempfile.mkdtemp()
    if ligand is not None:
        protein = "({0}) and not ({1})".format(protein, ligand)
        
        lig_file = os.path.join(out_dir, "lig.pdb")
        cmd.save(lig_file, ligand)
    else:
        lig_file = None
        
    prot_atoms = cmd.count_atoms(protein)
    if prot_atoms == 0:
        print("Error: no atoms included in protein selection")
        return
    elif prot_atoms < 50:
        print("Warning: only {0} atoms included in protein selection".format(prot_atoms))

    prot_file = os.path.join(out_dir, "prot.pdb")
    cmd.save(prot_file, protein)

    if (mode is None) and ((ligand is not None) or (pocket_coordinate is not None) or (residue is not None)):
        mode = "specific"
    elif mode is None:
        mode = "largest"

    spheres = identify.pocket(prot_file, mode=mode, lig_file=lig_file, coordinate=pocket_coordinate, min_rad=min_rad, max_rad=max_rad, lig_excl_rad=lig_excl_rad, lig_incl_rad=lig_incl_rad, subdivide=subdivide, minimum_volume=minimum_volume)

    if mode in ["specific", "largest"]:
        if not subdivide:
            print("Pocket Volume: {0} A^3".format(format(spheres[0].mesh.volume, '.2f')))
            pymol_utilities.display_spheres_object(spheres[0], spheres[0].name, state=1, color=color, alpha=alpha, mode=display_mode)
        else:
            print("Whole Pocket Volume: {0} A^3".format(format(spheres[0].mesh.volume, '.2f')))
            pymol_utilities.display_spheres_object(spheres[0], spheres[0].name, state=1, color=color, alpha=alpha, mode=display_mode)
            palette = construct_palette(max_value=(len(spheres) -1))
            for index, sps in enumerate(spheres[1:]):
                group = int(sps.g[0])
                print("{0} volume: {1} A^3".format(sps.name, format(sps.mesh.volume, '.2f')))
                pymol_utilities.display_spheres_object(sps, sps.name, state=1, color=palette[index], alpha=alpha, mode=display_mode)
    else:
        palette = pymol_utilities.construct_palette(max_value=len(spheres))
        for index, s in enumerate(spheres):
            print("{0} volume: {1} A^3".format(s.name, format(s.mesh.volume, '.2f')))
            pymol_utilities.display_spheres_object(s, s.name, state=1, color=palette[index], alpha=alpha, mode=display_mode)

    if output_dir is None:
        shutil.rmtree(out_dir)
    return




import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="bio-pyvol",
    version="1.0.6",
    description="a PyMOL plugin for visualization, comparison, and volume calculation for protein drug-binding sites",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/rhs2132/pyvol",
    author="Ryan Smith",
    author_email="rhydesmith@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        ],
    packages=["pyvol"],
    install_requires=[
        "biopython>=1.73",
        "numpy>=1.16.1",
        "pandas>=0.24.1",
        "scipy>=1.2.1",
        "scikit-learn>=0.20.2",
        "trimesh>=2.36.29",
        "configparser"
        ],
    entry_points={
        "console_scripts": [
            "pyvol=pyvol.__main__:main",
            ]
        },
    )
